class SQSError extends Error {
    constructor(message) {
        super(message);
        this.name = this.constructor.name;
    }
}
class TimeoutError extends Error {
    constructor(message = "Operation timed out.") {
        super(message);
        this.message = message;
        this.name = "TimeoutError";
    }
}
class StandardError extends Error {
    constructor(message = "An unexpected error occurred:") {
        super(message);
        this.message = message;
        this.name = "StandardError";
    }
}
/**
 * Checks if the error provided should be treated as a connection error.
 * @param err The error that was received.
 */
function isConnectionError(err) {
    if (err instanceof SQSError) {
        return (err.statusCode === 403 ||
            err.code === "CredentialsError" ||
            err.code === "UnknownEndpoint" ||
            err.code === "AWS.SimpleQueueService.NonExistentQueue" ||
            err.code === "CredentialsProviderError");
    }
    return false;
}
/**
 * Formats an AWSError the the SQSError type.
 * @param err The error object that was received.
 * @param message The message to send with the error.
 */
function toSQSError(err, message) {
    const sqsError = new SQSError(message);
    sqsError.code = err.name;
    sqsError.statusCode = err.$metadata?.httpStatusCode;
    sqsError.retryable = err.$retryable?.throttling;
    sqsError.service = err.$service;
    sqsError.fault = err.$fault;
    sqsError.time = new Date();
    return sqsError;
}
/**
 * Formats an Error to the StandardError type.
 * @param err The error object that was received.
 * @param message The message to send with the error.
 */
function toStandardError(err, message) {
    const error = new StandardError(message);
    error.cause = err;
    error.time = new Date();
    return error;
}
/**
 * Formats an Error to the TimeoutError type.
 * @param err The error object that was received.
 * @param message The message to send with the error.
 */
function toTimeoutError(err, message) {
    const error = new TimeoutError(message);
    error.cause = err;
    error.time = new Date();
    return error;
}
export { SQSError, TimeoutError, isConnectionError, toSQSError, toStandardError, toTimeoutError, };
